// This is LinkedList.h
//
// Copyright 1997, 1998        Mark Manca
//

#ifndef LINKEDLIST_H
#define LINKEDLIST_H


#include <stdlib.h>



template <class Type>
class LinkedListItem {
    public:
        LinkedListItem(Type& data,
                       LinkedListItem<Type> *next=NULL,
                       LinkedListItem<Type> *previous=NULL);
        LinkedListItem();
        ~LinkedListItem();


        Type getData()
            { return _data; }

        void  setData(Type& data)
            { _data = data; }


        LinkedListItem<Type> *getNextItem()
            { return _next; }

        void  setNextItem(LinkedListItem<Type> *item);


        LinkedListItem<Type> *getPreviousItem()
            { return _previous; }

        void  setPreviousItem(LinkedListItem<Type> *item);


        // template <class Type> friend class LinkedList;


    protected:
        Type _data;
        LinkedListItem<Type> *_next;
        LinkedListItem<Type> *_previous;


    private:
};





//
// Please Note: This data structure only does shallow copies (ie it just
// copies pointers around). So please make sure that the data inside of
// the list remains in scope for the life of the data structure.
//


template <class Type>
class LinkedList {
	public:
		LinkedList(Type& item);
		LinkedList();
		~LinkedList();

		//
		// return the head and the tail of the list. Move the current
		// item to be the head or tail of the list.
		//
		Type getHead();     
		Type getTail();

		Type getCurrentItem()
			{ return  _currentItem->getData(); }


		//
		// returns the next or previous item in the list, then moves
		// the current position to the item that was returned. Good for
		// traversals of the list.
		//
		Type getNextItem();
		Type getPreviousItem();


		//
		// Inserts. Moves current position to the item that was just inserted.
		//
		void insertItem(Type& item, int position);
		void insertNextItem(Type& item);
		void insertPreviousItem(Type& item);
		void appendItem(Type& item);
		void prependItem(Type& item);
	
		//
		// removes the specified item. current item becomes the item that takes
		// place of the removed item in the list.
		//
		void remove(int position);
		void removeHead();
		void removeTail();
		void removeCurrentItem();

		//
		// removes the item from the list. The current item remains the same.
		//
		void removeNextItem();
		void removePreviousItem();


		int getLength() 
			{ return _length; }


		//
		// access the linked list like it is an array.  This operator will
		// be smart enough to figure out the quickest way to the index from
		// the current index pointer's current position.
		//
		// if an index is given that is out of bounds, just return whatever
		// the current index pointer is pointing to.
		//
		Type operator[](int index);


		//
		// functions to traverse up and down the linked list
		//
		LinkedList& operator++();      // is prefix operator
		LinkedList& operator--();      // is prefix operator


	protected:
		LinkedListItem<Type> *_headOfList;
		LinkedListItem<Type> *_tailOfList;
		LinkedListItem<Type> *_currentItem;
		int _length;
		int _currentPosition;


	private:
};










//
// The whole impelementation of the linked list class has to be in
// this header file to satisfy the Microsoft compiler.  Don't you 
// just love Microsoft?
//



template <class Type>
LinkedList<Type>::LinkedList(Type& item) {
	_headOfList = new LinkedListItem<Type>(item, NULL, NULL);
	_tailOfList = _headOfList;
	_currentItem = _headOfList;

	_length = 1;
	_currentPosition = 0;
}


template <class Type>
LinkedList<Type>::LinkedList() {
	_headOfList = NULL;
	_tailOfList = NULL;
	_currentItem = NULL;
	_length = 0;
	_currentPosition = -1;
}



template <class Type>
LinkedList<Type>::~LinkedList() {
	for (int i=_length; i > 0; i--) {
		removeTail();
	}
}







//
// return the head and the tail of the list. Move the current
// item to be the head or tail of the list.
//
template <class Type>
Type LinkedList<Type>::getHead() {
	if (_headOfList) {
		_currentItem = _headOfList;
		_currentPosition = 0;

		return _headOfList->getData();
	}
	else {
		//
		// This is going to core, but it makes the compiler happy since we
		// are dealing with templates.
		//
		return _headOfList->getData();
	}
}



template <class Type>
Type LinkedList<Type>::getTail() {
	if (_tailOfList) {
		_currentItem = _tailOfList;
		_currentPosition = _length - 1;

		return _tailOfList->getData();
	}
	else {
		//
		// This is going to core, but it makes the compiler happy since we
		// are dealing with templates.
		//
		return _tailOfList->getData();
	}
}




//
// returns the next or previous item in the list, then moves
// the current position to the item that was returned. Good for
// traversals of the list.
//
template <class Type>
Type LinkedList<Type>::getNextItem() {
	if (_currentItem && _currentItem->getNextItem()) {
		_currentItem = _currentItem->getNextItem();
		_currentPosition++;

		return _currentItem->getData();
	}
	else {
		return NULL;
	}
}


template <class Type>
Type LinkedList<Type>::getPreviousItem() {
	if (_currentItem && _currentItem->getPreviousItem()) {
		_currentItem = _currentItem->getPreviousItem();
		_currentPosition--;

		return _currentItem->getData();
	}
	else {
		return NULL;
	}
}




//
// Inserts. Moves current position to the item that was just inserted.
//
template <class Type>
void LinkedList<Type>::insertItem(Type& item, int position) {
	if (position == _length) {
		appendItem(item);
	}
	else if (position < _length) {
		(*this)[position];
		insertPreviousItem(item);
	}
}



template <class Type>
void LinkedList<Type>::insertNextItem(Type& item) {
	LinkedListItem<Type> *newItem = new LinkedListItem<Type>(item, NULL, _currentItem);
	
	if (_currentItem && _currentItem->getNextItem()) {
		newItem->setNextItem(_currentItem->getNextItem());
		_currentItem->setNextItem(newItem);
	}

	_length++;

	if (!_headOfList) {
		_headOfList = newItem;
	}
	if (!_tailOfList) {
		_tailOfList = newItem;
	}
	_currentItem = newItem;
	_currentPosition++;
}



template <class Type>
void LinkedList<Type>::insertPreviousItem(Type& item) {
	LinkedListItem<Type> *newItem = new LinkedListItem<Type>(item, _currentItem, NULL);
	
	if (_currentItem && _currentItem->getPreviousItem()) {
		newItem->setPreviousItem(_currentItem->getPreviousItem());
		_currentItem->getPreviousItem()->setNextItem(newItem);
	}

	_length++;

	if (!_headOfList) {
		_headOfList = newItem;
	}
	if (!_tailOfList) {
		_tailOfList = newItem;
	}
	_currentItem = newItem;
}



template <class Type>
void LinkedList<Type>::appendItem(Type& item) {
	LinkedListItem<Type> *newItem = new LinkedListItem<Type>(item, NULL, _tailOfList);
	
	if (_tailOfList) {
		_tailOfList->setNextItem(newItem);
	}

	_length++;

	if (!_headOfList) {
		_headOfList = newItem;
	}
	_tailOfList = newItem;
	_currentItem = newItem;
	_currentPosition = _length - 1;
}



template <class Type>
void LinkedList<Type>::prependItem(Type& item) {
	LinkedListItem<Type> *newItem = new LinkedListItem<Type>(item, _headOfList, NULL);
	
	if (_headOfList) {
		_headOfList->setPreviousItem(newItem);
	}

	_length++;

	if (!_tailOfList) {
		_tailOfList = newItem;
	}
	_headOfList = newItem;
	_currentItem = newItem;
	_currentPosition = 0;
}




//
// removes the specified item. current item becomes the item that takes
// place of the removed item in the list.
//
template <class Type>
void LinkedList<Type>::remove(int position) {
	(*this)[position];
	removeCurrentItem();
}



template <class Type>
void LinkedList<Type>::removeHead() {
	if (_headOfList) {
		LinkedListItem<Type> *nextItem = _headOfList->getNextItem();

		_length--;
		_currentPosition=0;

		delete _headOfList;

		if (nextItem) {
			nextItem->setPreviousItem(NULL);
			_currentItem = nextItem;
			_headOfList = nextItem;
		}
		else {
			if (_headOfList == _tailOfList) {
				_tailOfList = NULL;
			}

			_currentItem = NULL;
			_headOfList = NULL;
			_currentPosition = -1;
		}
	}
}



template <class Type>
void LinkedList<Type>::removeTail() {
	if (_tailOfList) {
		LinkedListItem<Type> *previousItem = _tailOfList->getPreviousItem();

		_length--;
		_currentPosition = _length - 1;

		delete _tailOfList;

		if (previousItem) {
			previousItem->setNextItem(NULL);
			_currentItem = previousItem;
			_tailOfList = previousItem;
		}
		else {
			if (_tailOfList == _headOfList) {
				_headOfList = NULL;
			}

			_currentItem = NULL;
			_tailOfList = NULL;
		}

		_currentPosition--;
	}
}



template <class Type>
void LinkedList<Type>::removeCurrentItem() {
	if (_currentItem) {
		if (_currentItem->getPreviousItem() &&
			_currentItem->getNextItem()) {
			LinkedListItem<Type> *nextItem = _currentItem->getNextItem();

			_currentItem->getPreviousItem()->setNextItem(nextItem);
			nextItem->setPreviousItem(_currentItem->getPreviousItem());

			delete _currentItem;
			_currentItem = NULL;

			_currentItem = nextItem;
			_length--;
		}
		else if (_currentItem->getPreviousItem()) {
			removeTail();
		}
		else if (_currentItem->getNextItem()) {
			removeHead();
		}
		else {
			delete _currentItem;
			_currentItem = NULL;
			_headOfList = NULL;
			_tailOfList = NULL;
			_length = 0;
			_currentPosition = -1;
		}
	}
}



//
// removes the item from the list. The current item remains the same.
//
template <class Type>
void LinkedList<Type>::removeNextItem() {
	if (_currentItem && _currentItem->getNextItem()) {
		LinkedListItem<Type> *nextItem = _currentItem->getNextItem();

		_currentItem->setNextItem(nextItem->getNextItem());
		_currentItem->getNextItem()->setPreviousItem(_currentItem);

		if (_tailOfList == nextItem) {
			_tailOfList = _currentItem;
		}

		delete nextItem;
		nextItem = NULL;

		_length--;
	}
}



template <class Type>
void LinkedList<Type>::removePreviousItem() {
	if (_currentItem && _currentItem->getPreviousItem()) {
		LinkedListItem<Type> *previousItem = _currentItem->getPreviousItem();

		_currentItem->setPreviousItem(previousItem->getPreviousItem());
		_currentItem->getPreviousItem()->setNextItem(_currentItem);

		if (_headOfList == previousItem) {
			_headOfList = previousItem;
		}

		delete previousItem;
		previousItem = NULL;

		_length--;
		_currentPosition--;
	}
}



template <class Type>
Type LinkedList<Type>::operator[](int index) {
	if (_length > 0 && index >= 0 && index < _length) {
		int distance = index - _currentPosition;

		if (_currentPosition == index) {
			;
		}
		// index is closest to the currentPosition
		else if (abs(distance) <= index &&
		         abs(distance) < _length - index) {
			if (distance > 0) {
				while (_currentPosition != index) {
					++(*this);
				}
			}
			else {
				while (_currentPosition != index) {
					--(*this);
				}
			}
		}
		// we are closest to the tail
		else if (abs(distance) == index && 
		         abs(distance) >= _length - index) {
			_currentItem = _tailOfList;
			_currentPosition = _length - 1;

			while (_currentPosition != index) {
				--(*this);
			}
		}
		// index is closest to the head
		else if (abs(distance) > index &&
		         abs(distance) < _length - index) {
			_currentItem = _headOfList;
			_currentPosition = 0;

			while (_currentPosition != index) {
				++(*this);
			}
		}
		// index is closest to the tail 
		else {
			_currentItem = _tailOfList;
			_currentPosition = _length - 1;

			while (_currentPosition != index) {
				--(*this);
			}
		}

		return _currentItem->getData();
	}
	else {
		return _currentItem->getData();
	}
}



//
// prefix addition operator (increment where the list is)
//
template <class Type>
LinkedList<Type>& LinkedList<Type>::operator++() {
	if (_length > 0 && _currentPosition < _length - 1) {
		if (_currentItem->getNextItem()) {
			_currentItem = _currentItem->getNextItem();
			_currentPosition++;
		}
	}

	return *this;
}



//
// prefix subtraction operator (decrement where the current item 
// and current pos)
//
template <class Type>
LinkedList<Type>& LinkedList<Type>::operator--() {
	if (_length > 0 && _currentPosition > 0) {
		if (_currentItem->getPreviousItem()) {
			_currentItem = _currentItem->getPreviousItem();
			_currentPosition--;
		}
	}

	return *this;
}








template <class Type>
LinkedListItem<Type>::LinkedListItem(Type& data, 
							   LinkedListItem<Type> *next,
							   LinkedListItem<Type> *previous)
	: _data(data),
	  _next(next),
	  _previous(previous) {

	// empty
}



template <class Type>
LinkedListItem<Type>::LinkedListItem() 
	: _data(NULL),
	  _next(NULL),
	  _previous(NULL) {

	// empty
}



template <class Type>
LinkedListItem<Type>::~LinkedListItem() {
	// empty
}




template <class Type>
void  LinkedListItem<Type>::setNextItem(LinkedListItem<Type> *item) {
#if 0
	item->_previous = this;
	item->_next = _next;

	_next->_previous = item;
#endif

	_next = item;

	return;
}




template <class Type>
void  LinkedListItem<Type>::setPreviousItem(LinkedListItem<Type> *item) {
#if 0
	item->_previous = _previous;
	item->_next = this;

	_previous->_next = item;
#endif

	_previous = item;

	return;
}




#endif    // LINKEDLIST_H


