// This is Color.h
//
// Copyright 1997, 1998     Mark Manca
//

#ifndef _COLOR_H_
#define _COLOR_H_


#include <iostream.h>


class Color {
	public:
		enum CompositingOperator { A=0, B,
		                           AOVERB, BOVERA,
		                           AINB, BINA,
		                           AOUTB, BOUTA,
		                           AATOPB, BATOPA,
		                           AXORB };

		//
		// Constructors, destructors...
		//
		inline Color(unsigned char red=(unsigned char)255, 
		             unsigned char green=(unsigned char)255, 
		             unsigned char blue=(unsigned char)255, 
		             unsigned char alpha=(unsigned char)255);
		inline Color(const unsigned char* color);
		inline Color(const Color& color);
		inline ~Color();


		//
		// set and get various color attributes...
		//
		inline void setRed(unsigned char red);
		inline void setGreen(unsigned char green);
		inline void setBlue(unsigned char blue);
		inline void setAlpha(unsigned char alpha);

		inline void setRGB(unsigned char red, unsigned char green, unsigned char blue);
		inline void setRGB(const unsigned char *color);
		inline void setRGBA(unsigned char red, unsigned char green, unsigned char blue, unsigned char alpha);
		inline void setRGBA(const unsigned char *color);


		inline unsigned char getRed() const;
		inline unsigned char getGreen() const;
		inline unsigned char getBlue() const;
		inline unsigned char getAlpha() const;

		inline void getRGB(unsigned char *red, unsigned char *green, unsigned char *blue) const;
		inline const unsigned char *getRGB() const;
		inline void getRGBA(unsigned char *red, unsigned char *green, 
		                    unsigned char *blue, unsigned char *alpha) const;
		inline const unsigned char *getRGBA() const;


		inline unsigned char& operator[](int index);
		inline Color& operator=(const Color& rhs);
		inline int operator==(const Color& rhs);
		inline int operator!=(const Color& rhs);
		//inline ostream& operator<<(ostream& os, Color& color);


		//
		// compositing operators. All of the operators consider the object
		// that is acted upon to be a, and the argument to the function b.
		//
		// these compositing operators assume that the color has been 
		// been premultiplied by its respective alpha value.
		//
		// all of the compositing operators modify the color that is given
		// as the argument (b) with the resulting color of the compositing
		// operation in order to conserve memory and to increase speed.
		//
		inline Color& a(Color& color);
		inline Color& b(Color& color);
		inline Color& aoverb(Color& color);
		inline Color& bovera(Color& color);
		inline Color& ainb(Color& color);
		inline Color& bina(Color& color);
		inline Color& aoutb(Color& color);
		inline Color& bouta(Color& color);
		inline Color& aatopb(Color& color);
		inline Color& batopa(Color& color);
		inline Color& axorb(Color& color);

		//
		// This function is by far the most general as well as by far
		// the slowest.
		//
		inline Color& composite(Color& color, CompositingOperator compOperator);


	protected:
		unsigned char _color[4];


	private:
};




//
// Put all of the inline functions here because of the microsoft compiler
//

inline Color::Color(unsigned char red, unsigned char green, unsigned char blue, unsigned char alpha)  {
	_color[0] = red;
	_color[1] = green; 
	_color[2] = blue; 
	_color[3] = alpha;
}


inline Color::Color(const unsigned char* color) {
	_color[0] = color[0];
	_color[1] = color[1];
	_color[2] = color[2];
	_color[3] = color[3];
}


inline Color::Color(const Color& color) {
	_color[0] = color._color[0];
	_color[1] = color._color[1];
	_color[2] = color._color[2];
	_color[3] = color._color[3];
}


inline Color::~Color() {
	// empty
}




//
// set and get various color attributes...
//
inline void Color::setRed(unsigned char red) {
	_color[0] = red;
}


inline void Color::setGreen(unsigned char green) {
	_color[1] = green;
}


inline void Color::setBlue(unsigned char blue) {
	_color[2] = blue;
}


inline void Color::setAlpha(unsigned char alpha) {
	_color[3] = alpha;
}



inline void Color::setRGB(unsigned char red, unsigned char green, unsigned char blue) {
	_color[0] = red;
	_color[1] = green;
	_color[2] = blue;
}


inline void Color::setRGB(const unsigned char *color) {
	_color[0] = color[0];
	_color[1] = color[1];
	_color[2] = color[2];
}


inline void Color::setRGBA(unsigned char red, unsigned char green, unsigned char blue, unsigned char alpha) {
	_color[0] = red;
	_color[1] = green;
	_color[2] = blue;
	_color[3] = alpha;
}


inline void Color::setRGBA(const unsigned char *color) {
	_color[0] = color[0];
	_color[1] = color[1];
	_color[2] = color[2];
	_color[3] = color[3];
}



inline unsigned char Color::getRed() const {
	return _color[0];
}


inline unsigned char Color::getGreen() const {
	return _color[1];
}


inline unsigned char Color::getBlue() const {
	return _color[2];
}


inline unsigned char Color::getAlpha() const {
	return _color[3];
}



inline void Color::getRGB(unsigned char *red, unsigned char *green, unsigned char *blue) const {
	*red   = _color[0];
	*green = _color[1];
	*blue  = _color[2];
}


inline const unsigned char* Color::getRGB() const {
	return _color;
}


inline void Color::getRGBA(unsigned char *red, unsigned char *green, 
                           unsigned char *blue, unsigned char *alpha) const {
	*red   = _color[0];
	*green = _color[1];
	*blue  = _color[2];
	*alpha = _color[3];
}


inline const unsigned char* Color::getRGBA() const {
	return _color;
}




inline unsigned char& Color::operator[](int index) {
	if (!(index >= 0 && index < 4)) {
		cerr << "Warning: color array index: " << index 
		     << " is out of bounds" << endl;
	}

	return _color[index];
}


inline Color& Color::operator=(const Color& rhs) {
	_color[0] = rhs._color[0];
	_color[1] = rhs._color[1];
	_color[2] = rhs._color[2];
	_color[3] = rhs._color[3];

	return *this;
}


inline int Color::operator==(const Color& rhs) {
	return (_color[0] == rhs._color[0] && _color[1] == rhs._color[1] &&
	        _color[2] == rhs._color[2] && _color[3] == rhs._color[3]);
}


inline int Color::operator!=(const Color& rhs) {
	return !(*this == rhs);
}



inline ostream& operator<<(ostream& os, Color& color) {
	Color aColor = color;
	os << (int)aColor[0] << " " << (int)aColor[1] << " "
	   << (int)aColor[2] << " " << (int)aColor[3];

	return os;
}






//
// compositing operators. All of the operators consider the object
// that is acted upon to be a, and the argument to the function b.
//
// all of the compositing operators modify the color that is given
// as the argument (b) with the resulting color of the compositing
// operation in order to conserve memory and to increase speed.
//
//
// compositing algebra:
//
//      c  = F c  + F c
//       o    A A    B B
//

//
// Fa = 1;   Fb = 0;
//
inline Color& Color::a(Color& color) {
	color._color[0] = _color[0];
	color._color[1] = _color[1];
	color._color[2] = _color[2];
	color._color[3] = _color[3];

	return *this;
}


//
// Fa = 0; Fb = 1;
//
inline Color& Color::b(Color& /*color*/) {
	//
	// don't do anything here because A gets ignored
	//

	return *this;
}




//
// Fa = 1;    Fb = 1-alphaA
//
inline Color& Color::aoverb(Color& color) {
	color._color[0] = (unsigned char)((float)_color[0] + 
	                  ((255-_color[3])/255.0f) * (float)color._color[0]);
	color._color[1] = (unsigned char)((float)_color[1] + 
	                  ((255-_color[3])/255.0f) * (float)color._color[1]);
	color._color[2] = (unsigned char)((float)_color[2] + 
	                  ((255-_color[3])/255.0f) * (float)color._color[2]);
	color._color[3] = (unsigned char)((float)_color[3] + 
	                  ((255-_color[3])/255.0f) * (float)color._color[3]);

	return *this;
}



//
// Fa = 1-alphaB;    Fb = 1;
//
inline Color& Color::bovera(Color& color) {
	color._color[0] = (unsigned char)((float)_color[0] * ((255-color._color[3])/255.0f)
	                  + (float)color._color[0]);
	color._color[1] = (unsigned char)((float)_color[1] * ((255-color._color[3])/255.0f)
	                  + (float)color._color[1]);
	color._color[2] = (unsigned char)((float)_color[2] * ((255-color._color[3])/255.0f)
	                  + (float)color._color[2]);
	color._color[3] = (unsigned char)((float)_color[3] * ((255-color._color[3])/255.0f)
	                  + (float)color._color[3]);

	return *this;
}



//
// Fa = alphaB, Fb = 0
//
inline Color& Color::ainb(Color& color) {
	color._color[0] = 
		(unsigned char)((float)_color[0] * ((float)color._color[3]/255.0f));
	color._color[1] = 
		(unsigned char)((float)_color[1] * ((float)color._color[3]/255.0f));
	color._color[2] = 
		(unsigned char)((float)_color[2] * ((float)color._color[3]/255.0f));
	color._color[3] = 
		(unsigned char)((float)_color[3] * ((float)color._color[3]/255.0f));

	return *this;
}



//
// Fa = 0, Fb = alphaA
//
inline Color& Color::bina(Color& color) {
	color._color[0] = 
		(unsigned char)((float)color._color[0] * ((float)_color[3]/255.0f));
	color._color[1] = 
		(unsigned char)((float)color._color[1] * ((float)_color[3]/255.0f));
	color._color[2] = 
		(unsigned char)((float)color._color[2] * ((float)_color[3]/255.0f));
	color._color[3] = 
		(unsigned char)((float)color._color[3] * ((float)_color[3]/255.0f));

	return *this;
}



//
// Fa = 1 - alphaB, Fb = 0
//
inline Color& Color::aoutb(Color& color) {
	color._color[0] = (unsigned char)((float)_color[0] * ((255-color._color[3])/255.0f));
	color._color[1] = (unsigned char)((float)_color[1] * ((255-color._color[3])/255.0f));
	color._color[2] = (unsigned char)((float)_color[2] * ((255-color._color[3])/255.0f));
	color._color[3] = (unsigned char)((float)_color[3] * ((255-color._color[3])/255.0f));

	return *this;
}



//
// Fa = 0,   Fb = 1 - alphaA
//
inline Color& Color::bouta(Color& color) {
	color._color[0] = (unsigned char)((float)color._color[0] * ((255-_color[3])/255.0f));
	color._color[1] = (unsigned char)((float)color._color[1] * ((255-_color[3])/255.0f));
	color._color[2] = (unsigned char)((float)color._color[2] * ((255-_color[3])/255.0f));
	color._color[3] = (unsigned char)((float)color._color[3] * ((255-_color[3])/255.0f));

	return *this;
}



//
// Fa = alphaB,   Fb = 1 - alphaA
//
inline Color& Color::aatopb(Color& color) {
	color._color[0] = 
		(unsigned char)(((float)_color[0] * ((float)color._color[3]/255.0f))
	           + ((float)color._color[0] * ((255-_color[3])/255.0f)));
	color._color[1] =
		(unsigned char)(((float)_color[1] * ((float)color._color[3]/255.0f))
		       + ((float)color._color[1] * ((255-_color[3])/255.0f)));
	color._color[2] =
		(unsigned char)(((float)_color[2] * ((float)color._color[3]/255.0f))
		       + ((float)color._color[2] * ((255-_color[3])/255.0f)));
	color._color[3] =
		(unsigned char)(((float)_color[3] * ((float)color._color[3]/255.0f))
		       + ((float)color._color[3] * ((255-_color[3])/255.0f)));

	return *this;
}



//
// Fa = 1 - alphaB,   Fb = alphaA
//
inline Color& Color::batopa(Color& color) {
	color._color[0] =
		(unsigned char)(((float)_color[0] * ((255 - color._color[3])/255.0f))
               + ((float)color._color[0] * ((float)_color[3]/255.0f)));
	color._color[1] =
		(unsigned char)(((float)_color[1] * ((255 - color._color[3])/255.0f))
		       + ((float)color._color[1] * ((float)_color[3]/255.0f)));
	color._color[2] =
		(unsigned char)(((float)_color[2] * ((255 - color._color[3])/255.0f))
		       + ((float)color._color[2] * ((float)_color[3]/255.0f)));
	color._color[3] =
		(unsigned char)(((float)_color[3] * ((255 - color._color[3])/255.0f))
		       + ((float)color._color[3] * ((float)_color[3]/255.0f)));

	return *this;
}



//
// Fa = 1-alphaB,    Fb = 1-alphaA
//
inline Color& Color::axorb(Color& color) {
	color._color[0] = 
		(unsigned char)(((float)_color[0] * ((255 - color._color[3])/255.0f))
		       + ((float)color._color[0] * ((255 - _color[3])/255.0f)));
	color._color[1] =
		(unsigned char)(((float)_color[1] * ((255 - color._color[3])/255.0f))
		       + ((float)color._color[1] * ((255 - _color[3])/255.0f)));
	color._color[2] =
		(unsigned char)(((float)_color[2] * ((255 - color._color[3])/255.0f))
		       + ((float)color._color[2] * ((255 - _color[3])/255.0f)));
	color._color[3] =
		(unsigned char)(((float)_color[3] * ((255 - color._color[3])/255.0f))
		       + ((float)color._color[3] * ((255 - _color[3])/255.0f)));

	return *this;
}




//
// This function is by far the most general as well as by far
// the slowest.
//
inline Color& Color::composite(Color& color, 
                               CompositingOperator compOperator) {
	switch (compOperator) {
		case A:
			return a(color);
		case B:
			return b(color);
		case AOVERB:
			return aoverb(color);
		case BOVERA:
			return bovera(color);
		case AINB:
			return ainb(color);
		case BINA:
			return bina(color);
		case AOUTB:
			return aoutb(color);
		case BOUTA:
			return bouta(color);
		case AATOPB:
			return aatopb(color);
		case BATOPA:
			return batopa(color);
		case AXORB:
			return axorb(color);
		default:
			break;
	}

	return *this;
}


#endif    // _COLOR_H_
