// This is ImageFileGIFLoad.cpp
//
// Copyright 1997, 1998     Mark Manca
//

// This file contains the code to load a gif image

#ifdef WIN32
#include <windows.h>
#endif

#include <GL/gl.h>
#include <string.h>
#include <stdio.h>

#include "ImageFile.h"
#include "gifFileHeader.h"

#ifdef BIG_ENDIAN
#include "byteSwap.h"
#endif



// These functions assumes that the file to be read is already open

int ImageFile::_readGIF87aHeader() {
	gifFileHeader fileInfo;

	return _readGIF87aHeader(&fileInfo);
}




int ImageFile::_readGIF87aHeader(gifFileHeader *fileInfo) {
	int globalColorMapFollows;
	int numBitsOfColor;
	int numBitsPerPixel;
	int status;


	if (!__imageFile || !__isOpen) {
    	return 0;
	}

	//
	// get to the beginning of the file so we can read the 
	// header into the struct
	//
	status = fseek(__imageFile, 0, SEEK_SET);
	if (status) {
		return 0;			// an error occured
	}


	//
	// read in the header into the data structure
	//
	status = fread((char *)fileInfo, 1, sizeof(gifFileHeader), __imageFile);

	// should we use ferror here??
	if (status != sizeof(gifFileHeader)) {
		return 0;
	}


	//
	// All gif images use color indexing and have at most 256 color entries
	// in there color tables. So, lets set some values now...
	//

	_imageType = GL_COLOR_INDEX;
	_imageDataType = GL_UNSIGNED_BYTE;
	_numChannels = 1;
	_colormapDataType = GL_UNSIGNED_BYTE;


	// bytes need to be swapped on sgis and other unix systems...
	#ifdef BIG_ENDIAN
		fileInfo->screenWidth = byteSwap(fileInfo->screenWidth);
		fileInfo->screenHeight = byteSwap(fileInfo->screenHeight);
	#endif

	globalColorMapFollows = fileInfo->additionalImageColorInfo & 0x80;
	numBitsOfColor = ((fileInfo->additionalImageColorInfo & 0x70) >> 4) + 1;
	numBitsPerPixel = (fileInfo->additionalImageColorInfo & 0x07) + 1;
	_colormapSize = 1 << ((int)numBitsPerPixel);
	_width = fileInfo->screenWidth;
	_height = fileInfo->screenHeight;


	#ifdef _DEBUG
		printf("Screen width: %d\nScreen height: %d\n",
			   fileInfo->screenWidth, fileInfo->screenHeight);

		if (globalColorMapFollows)	
			printf("global Color Map follows descriptor\n");
		else
			printf("global Color Map does not follow descriptor\n");

		printf("Number of bits of color resolution: %d\n", numBitsOfColor);
		printf("Bits per pixel in image: %d\n", numBitsPerPixel);
		printf("index of screen background color: %d\n", (int)fileInfo->backgroundColorIndex);
		printf("placeholder: %d\n", (int)fileInfo->placeholder);
		fflush(stdout);
	#endif

	return 1;
}





int ImageFile::_readGIF87aFile() {
	gifFileHeader fileInfo;

	if (!_readGIF87aHeader(&fileInfo)) {
		return 0;
	}


	//
	// allocate the memory that will be used to store the image
	//
	_allocateImageData();


	//
	// allocate the color map that will be necessary to use since
	// all gif images are color mapped.
	//
	_allocateColormapData();


	int globalColorMapFollows = fileInfo.additionalImageColorInfo & 0x80;
	int status;


	//
	// read in the global color map from the file
	//
	if (globalColorMapFollows) {

		//
		// read in the colormap
		//
		status = fread(_colormapData, 1, _colormapSize*3, __imageFile);

		// should we use ferror here??
		if (status != _colormapSize*3) {
			return 0;
		}
	}


	int imageInterlaced;
	int useLocalColorMap;
	int numBitsPerPixel;
	int numBitsOfColor;

	gifImageHeader imageInfo;


	numBitsOfColor = ((fileInfo.additionalImageColorInfo & 0x70) >> 4) + 1;


	while (1) {
		char byte='a';
		unsigned char byteCount;

		status = fread(&byte, 1, 1, __imageFile);
		if (status != 1) {
			return 0;
		}

		status = fseek(__imageFile, -1, SEEK_CUR);
		if (status) {
			return 0;			// an error occured
		}


		//
		// now that we have read a byte out of the file, we have to 
		// figure out what is coming so we can handle it correctly
		//
		if (byte != ',' && byte != '!') {
			break;
		}


		//
		// handle any GIF extensions that come along. Right now we don't
		// know how to handle any of them, so skip all of them...
		//

		if (byte == '!') {
			char garbage[256];

			//
			// reread the '!'
			//
			status = fread(&byte, 1, 1, __imageFile);
			if (status != 1) {
				return 0;
			}

			//
			// read the function code and ignore it
			//
			status = fread(&byte, 1, 1, __imageFile);
			if (status != 1) {
				return 0;
			}


			//
			// read the byte count of extension data that follows
			//
			status = fread(&byteCount, 1, 1, __imageFile);
			if (status != 1) {
				return 0;
			}

			while (byteCount) {
				//
				// read the extension specific data
				//
				status = fread(garbage, 1, byteCount, __imageFile);
				if (status != byteCount) {
					return 0;
				}

				//
				// read the byte count of extension data that follows
				//
				status = fread(&byteCount, 1, 1, __imageFile);
				if (status != 1) {
					return 0;
				}
			}
			continue;
		}



		//
		// if we are down here, we have an image....
		//

		//
		// read in the header into the data structure
		//
		status = fread((char *)&imageInfo, 1, sizeof(gifImageHeader), __imageFile);

		// should we use ferror here??
		if (status != sizeof(gifImageHeader)) {
			return 0;
		}


		#ifdef BIG_ENDIAN
			imageInfo.imageLeft = byteSwap(imageInfo.imageLeft);
			imageInfo.imageTop = byteSwap(imageInfo.imageTop);
			imageInfo.imageWidth = byteSwap(imageInfo.imageWidth);
			imageInfo.imageHeight = byteSwap(imageInfo.imageHeight);
		#endif


		useLocalColorMap = imageInfo.additionalImageColorInfo & 0x80;
		imageInterlaced = imageInfo.additionalImageColorInfo & 0x40;

		if (useLocalColorMap) {
			numBitsPerPixel = (imageInfo.additionalImageColorInfo & 0x07) + 1;
		}
		else {
			numBitsPerPixel = (fileInfo.additionalImageColorInfo & 0x07) + 1;
		}


		#ifdef _DEBUG
			printf("\nImage Left: %d\n", imageInfo.imageLeft);
			printf("Image Top: %d\n", imageInfo.imageTop);
			printf("Image width: %d\n", imageInfo.imageWidth);
			printf("Image height: %d\n", imageInfo.imageHeight);

			if (useLocalColorMap)
				printf("should use local color map\n");
			else
				printf("should use global color map\n");

			if (imageInterlaced)
				printf("image formatted in interlaced order\n");
			else
				printf("image formatted in sequential order\n");

			printf("Bits per pixel in image: %d\n", numBitsPerPixel);
			fflush(stdout);
		#endif


		// read in the global color map from the file
		if (useLocalColorMap) {
			_colormapSize = 1 << ((int)numBitsPerPixel);

			//
			// allocate the color map that will be necessary to use since
			// all gif images are color mapped.
			//
			_allocateColormapData();


			//
			// read in the colormap
			//
			status = fread(_colormapData, 1, _colormapSize*3, __imageFile);

			// should we use ferror here??
			if (status != _colormapSize*3) {
				return 0;
			}
		}



		//
		// now we are ready to actually begin reading in the image data...
		//
		char codeSize;

		//
		// read the function code and ignore it
		//
		status = fread(&codeSize, 1, 1, __imageFile);
		if (status != 1) {
			return 0;
		}


		//
		// read the byte count of extension data that follows
		//
		status = fread(&byteCount, 1, 1, __imageFile);
		if (status != 1) {
			return 0;
		}

		// this is temp
		char garbage2[256];

		while (byteCount) {
			//
			// read the extension specific data
			//
			status = fread(garbage2, 1, byteCount, __imageFile);
			if (status != byteCount) {
				return 0;
			}

			//
			// read the byte count of extension data that follows
			//
			status = fread(&byteCount, 1, 1, __imageFile);
			if (status != 1) {
				return 0;
			}
		}


		//
		// this is temp
		//
		unsigned char* imageData = (unsigned char*)_imageData;
		unsigned char* imageDataPtr = imageData;//+(_width+_extraRowWidth)*_height;

		//
		// we're going to have to switch things around because
		// in gif, the origin starts at the upper left hand corner
		//
#if 0
		imageDataPtr -= _width+_extraRowWidth;
		imageDataPtr -= (_height-(imageInfo.imageTop+imageInfo.imageHeight)) *
		                (_width+_extraRowWidth);
#endif

		for (int i=imageInfo.imageTop; 
		     i < imageInfo.imageTop+imageInfo.imageHeight;
			 i++) {
				imageDataPtr += imageInfo.imageLeft;

			for (int j=imageInfo.imageLeft;
			     j < imageInfo.imageLeft+imageInfo.imageWidth;
				 j++) {
				*imageDataPtr++ = j%256;
			}
			imageDataPtr = imageData + (_width+_extraRowWidth)*(i+1);
#if 0
			imageDataPtr -= imageInfo.imageLeft + imageInfo.imageWidth;
			imageDataPtr -= _width+_extraRowWidth;
#endif
		}
	}



	char gifTerminator;

	status = fread(&gifTerminator, 1, 1, __imageFile);

	// should we use ferror here??
	if (status != 1) {
		return 0;
	}


	if (gifTerminator != ';') {
		//fprintf(stderr, "GIF Terminator not found!!\n");
		return 0;
	}


	return 1;
}

