// This is ImageFileGIFWrite.cpp

//
// Copyright 1997, 1998     Mark Manca
//

#ifdef WIN32
#include <windows.h>
#endif

#include <GL/gl.h>
#include <string.h>
#include <stdio.h>

#include "../DataStructures/OpenHashTable.h"

#include "ImageFile.h"
#include "gifFileHeader.h"


#ifdef BIG_ENDIAN
#include "byteSwap.h"
#endif


#ifndef MIN
#define MIN(a,b) ((a)<(b)?(a):(b))
#endif
#ifndef MAX
#define MAX(a,b) ((a)>(b)?(a):(b))
#endif



struct CodeTableEntry {
	unsigned char* string;
	int stringLen;
	int code;
	int bitsInCode;
};


static void initializeCodeTable(OpenHashTable<CodeTableEntry*>* hashCodeTable, 
                                CodeTableEntry* codeTable,
						        unsigned char* initialVals, int numInitVals,
							    unsigned int clearCode, int* bitsPerCode);


//
// see if the current prefix + the current pixel is in the table.
//
static int isInTable(unsigned char* currentPrefix, int currentPrefixLen, 
					 unsigned char currentPixel, 
					 OpenHashTable<CodeTableEntry*>* hashCodeTable,
					 int lastCodeTableEntry);



//
// return the code that is contained in the table.  The bitsInCode will
// get set to the number of bits in the code that are significant.
//
// if you use this function, the prefix must be in the table or else you 
// will most likely core or go into an infinite loop.  Use isInTable to
// see if the string is in the table.
//
static unsigned int lookUpCodeInTable(unsigned char* currentPrefix, 
									  int currentPrefixLen, 
									  OpenHashTable<CodeTableEntry*>* hashCodeTable,
									  int* bitsInCode);



//
// add the specified prefix to the table.  A new code will automatically
// be assigned.  lastCodeTableEntry will be incremented if necessary.
// If the code was successfully added, this function will return 1.  If the
// table is full, it will return 0.
//
static int addPrefixToCodeTable(unsigned char* prefix, int prefixLen, 
						        OpenHashTable<CodeTableEntry*>* hashCodeTable, 
								CodeTableEntry* codeTable,
						        int* lastCodeTableEntry);



//
// find the number of significant bits that the given word contains.
//
static int numSignificantBits(unsigned int word);


//
// this function will add the given code to the data bytes.
// It has a side effect of updating the currentDataBit variable.
//
// it will return the total number of valid bytes in this block that
// have been written.
//
// This function will also write the dataBytes out to disk if the number
// of data bytes goes over the maximum allowable block size (255)
//
static int setDataBytes(unsigned int code, int numSigBits,
	                    unsigned char *dataBytes, int *currentDataBit,
						FILE* imageFile);



//
// write the specified block out to the file as specified by the gif format
// 
static int writeBlockToFile(unsigned char* dataBytes, 
							unsigned char blockByteCount,
							FILE* imageFile);








//
// we are assuming that the user has already converted this image to a color
// indexed image.  If they haven't, the function will immediately return.
//

int ImageFile::_writeGIF87aFile() {
	gifFileHeader header;
	int i, j;
	int maxColors, extraColorMapEntries;
	int status;


	if (!__imageFile || !__isOpen) {
		return 0;
	}


	if (_imageType != GL_COLOR_INDEX || !_imageData ||
	    (_colormapDataType != GL_BYTE &&
	      _colormapDataType != GL_UNSIGNED_BYTE)) {
		return 0;
	}



	//
	// now that we have done some sanity checks, write out the header.
	//

	strncpy(header.gifSignature, "GIF87a", 6);
	header.screenWidth = (unsigned short)_width;
	header.screenHeight = (unsigned short)_height;
	header.additionalImageColorInfo = 0;

	//
	// use global color map
	//
	header.additionalImageColorInfo |= (unsigned char)0x80;

	//
	// set the max number of colors in the image
	//
	int lastBit = numSignificantBits((unsigned int)_colormapSize);

	if (1 << lastBit-1 == _colormapSize) {
		// 
		// if we have an even power of 2, don't do anything
		//
	}
	else {
		//
		// we don't have an even power of 2
		//
		lastBit += 1;
	}

	maxColors = 1 << (lastBit-1);
	extraColorMapEntries = maxColors - _colormapSize;

	header.additionalImageColorInfo |= (unsigned char)(lastBit-2);


	//
	// set the number of bits of color resolution.  Set this to max since
	// I'm not really sure what it should be.  The documentation isn't clear
	// on it at all.
	//
	header.additionalImageColorInfo |= (unsigned char)0x70;


	//
	// fill in the rest of the header structure.
	//
	header.backgroundColorIndex = (unsigned char)0;
	header.placeholder = (unsigned char)0;

	#ifdef BIG_ENDIAN
		header.screenWidth = byteSwap(header.screenWidth);
		header.screenHeight = byteSwap(header.screenHeight);
	#endif


	//
	// actually write the header out to the file
	//
	status = fwrite((void*)&header, 1, sizeof(gifFileHeader), __imageFile);

	if (status != sizeof(gifFileHeader)) {
		return 0;
	}



	//
	// now that we are done writing the header out, we have to write the
	// color table entries out.
	//
	status = fwrite(_colormapData, 1, _colormapSize*3, __imageFile);

	if (status != _colormapSize*3) {
		return 0;
	}

	char extraColor[3];
	extraColor[0] = (char)0;
	extraColor[1] = (char)0;
	extraColor[2] = (char)0;

	//
	// write out any extra color table entries necessary since there have to
	// be an even power of 2 present in the file.
	//
	for (i=0; i < extraColorMapEntries; i++) {
		status = fwrite((void*)extraColor, 1, 3, __imageFile);

		if (status != 3) {
			return 0;
		}
	}





	//
	// now that we are done with the front matter, we can get down to the
	// guts of things.  First we have to write out the image header.
	//

	gifImageHeader imageHeader;

	imageHeader.imageSeparator = ',';
	imageHeader.imageLeft = 0;
	imageHeader.imageTop = 0;
	imageHeader.imageWidth = (unsigned short)_width;
	imageHeader.imageHeight = (unsigned short)_height;

	//
	// we want to use the global colormap, so ignore the pixel val that
	// is stored here.  We will also write out the image data sequentially,
	// so set that bit to 0 as well.
	//
	imageHeader.additionalImageColorInfo = 0;


	#ifdef BIG_ENDIAN
		imageHeader.imageWidth = byteSwap(imageHeader.imageWidth);
		imageHeader.imageHeight = byteSwap(imageHeader.imageHeight);
	#endif


	//
	// the image header is ready to be written out the the file, so write it.
	//
	status = fwrite((void*)&imageHeader, 1, sizeof(gifImageHeader), 
	                __imageFile);

	if (status != sizeof(gifImageHeader)) {
		return 0;
	}





	//
	// we are finally ready to actually write out the image data.  We will
	// use a version of the LZW compression algorithm as defined by the 
	// gif standard.
	//

	unsigned char codeSize = (unsigned char)lastBit-1;


	//
	// write the code size out to the file.  It represents the minimum number
	// of bits to represent a single code.  Note that this is the minimun
	// number.  It is usually equal to the number of colors in the image.
	// As we compress the image, the code size will dynamically increase
	// until we hit our 12 bit limit as defined by the gif standard.
	//
	status = fwrite(&codeSize, 1, 1, __imageFile);

	if (status != 1) {
		return 0;
	}


	//
	// the maximum code table will be 4096, so we will make a hash table that is
	// roughly double the size.  The max key length is just a guess.
	//
	OpenHashTable<CodeTableEntry*> hashCodeTable(10, 10007);
	CodeTableEntry codeTable[0xfff];
	unsigned char initialVals[256];

	int lastCodeTableEntry;
	int bitsInCode;


	//
	// initialize some basic things
	//
	for (i=0; i < 256; i++) {
		initialVals[i] = (unsigned char)i;
	}



	//
	// we are going to have at most 256 data bytes because the byteCount
	// number is only a byte long.
	//
	unsigned char blockByteCount = 0;
	unsigned char dataBytes[256];
	int currentDataBit=0;

	//
	// maxColors is an even power of 2
	//
	unsigned int clearCode = maxColors;

	initializeCodeTable(&hashCodeTable, codeTable, initialVals, _colormapSize, 
	                    clearCode, &bitsInCode);


	//
	// After the code table has been initialized, the first few code table 
	// entries contain all of the basic entries (0...nColors) mapped.  It
	// will also contain entries for the clear code and the end of information
	// code, so add 2 to the colormap size and call that the 
	// lastCodeTableEntry
	//
	lastCodeTableEntry = _colormapSize+2;

	for (i=0; i < 256; i++) {
		dataBytes[i] = 0;
	}


	bitsInCode = numSignificantBits(clearCode);

	
	// 
	// the first byte in the stream must be the clear code
	//
	blockByteCount = setDataBytes(clearCode, bitsInCode, dataBytes,
	                              &currentDataBit, __imageFile);


	int totalImageBytesWritten=0;
	int totalImageBytes = _width*_height;


	//
	// because the origin for GIF images is in the upper left hand corner
	// and the origin of the data that we are storing is in the lower right
	// hand corner (and because we have some extra bytes for efficiency
	// reasons), we will have to copy data into an array as we go in order
	// to make aspects of the compression simpler.  This wastes memory.
	// Oh well.
	//
	unsigned char* origImageData = new unsigned char[_width*_height];

	for (i=_height-1, j=0; i >= 0; i--, j++) {
		memcpy(origImageData + j*_width, 
		       ((unsigned char*)_imageData) + i*(_width+_extraRowWidth), _width);
	}

	unsigned char* imageData = origImageData;


	unsigned char* currentPrefix = imageData;
	int currentPrefixLen=0;

	unsigned int code;


	while (totalImageBytesWritten < totalImageBytes) {

		//
		// if the prefix plus the current pixel is in the table, continue
		// looping until we find a string that is not in the table.  
		//
		while (isInTable(currentPrefix, currentPrefixLen, *imageData,
		                 &hashCodeTable, lastCodeTableEntry)) {
			//totalImageBytesWritten++;
			currentPrefixLen++;
			imageData++;

			if (totalImageBytesWritten == totalImageBytes) {
				break;
			}
		}



		//
		// Once we have found something that is not in the table, we can 
		// output the table entry for the current prefix.  We then add the 
		// current prefix plus the current pixel value to the table for future
		// searches.  We then set the current prefix = to the current pixel
		// and start over again.
		// 

		//
		// look up the code for the current prefix in the string table.  
		// bitsPerCode will get set to the number of bits that are 
		// significant of the code.
		//
		code = lookUpCodeInTable(currentPrefix, currentPrefixLen, 
		                         &hashCodeTable, &bitsInCode);



		// 
		// put the code in the data stream
		//
		blockByteCount = setDataBytes(code, bitsInCode, dataBytes,
		                              &currentDataBit, __imageFile);

		if (totalImageBytesWritten == totalImageBytes) {
			break;
		}



		//
		// after we have written the code to the data stream, we can add
		// the current prefix + the current pixel to the table for future
		// lookups.
		//
		currentPrefixLen++;

		if (!addPrefixToCodeTable(currentPrefix, currentPrefixLen, 
		                          &hashCodeTable, codeTable,
								  &lastCodeTableEntry)) {

			//
			// if we failed to add the string to the table, the table was
			// full.  So, reinitialize it and write out the clear code
			// to the data stream.
			//
			initializeCodeTable(&hashCodeTable, codeTable, initialVals, 
			                    _colormapSize, clearCode, &bitsInCode);


			//
			// After the code table has been initialized, the first few code 
			// table entries contain all of the basic entries (0...nColors) 
			// mapped.  It will also contain entries for the clear code and 
			// the end of information code, so add 2 to the colormap size and 
			// call that the lastCodeTableEntry
			//
			lastCodeTableEntry = _colormapSize+2;



			// 
			// the first byte in the stream must be the clear code
			//
			blockByteCount = setDataBytes(clearCode, 12, dataBytes, 
										  &currentDataBit, __imageFile);
		}



		//
		// now that we have put the code in the dataBytes, we can reset the
		// code to the current pixel so we can start the process over
		//
		totalImageBytesWritten += currentPrefixLen - 1;

		currentPrefix = imageData;
		currentPrefixLen = 1;

		imageData++;
	}


	delete [] origImageData;


	//
	// now that we have written out the image data in the proper compressed
	// gif LZW format, we should write the End of Information Code.
	//

	// 
	// put the End of Information Code in the data stream
	//
	code = clearCode + 1;
	blockByteCount = setDataBytes(code, bitsInCode, dataBytes, 
	                              &currentDataBit, __imageFile);


	//
	// write it out to the file
	//
	if (blockByteCount && !writeBlockToFile(dataBytes, blockByteCount, 
	                                        __imageFile)) {
		return 0;
	}


	//
	// write out a zero length block out to signify termination of the
	// image.
	//
	blockByteCount = 0;
	if (!writeBlockToFile(dataBytes, blockByteCount, __imageFile)) {
		return 0;
	}






	//
	// we have finally written out the image in the gif file format.  Now
	// all we have to do is write out the gif terminator...
	//

	char terminator = ';';

	status = fwrite((void*)&terminator, 1, 1, __imageFile);

	if (status != 1) {
		return 0;
	}



	//
	// return success
	//

	return 1;
}












static void initializeCodeTable(OpenHashTable<CodeTableEntry*>* hashCodeTable, 
								CodeTableEntry* codeTable,
						        unsigned char* initialVals, 
								int numInitVals, unsigned int clearCode,
								int* bitsInCode) {

	int i;


	hashCodeTable->resetHashTable();


	*bitsInCode = numSignificantBits((unsigned int)numInitVals);
	//clearCode += 2;         // this becomes the 1st avail compression code

	//
	// we initialize the first codeTable entries to contain an entry for
	// each color table entry (ie the number of different colors in the image
	//
	for (i=0; i < numInitVals; i++) {
		codeTable[i].string = &(initialVals[i]);
		codeTable[i].stringLen = 1;
		codeTable[i].code = i; // + clearCode;
		codeTable[i].bitsInCode = *bitsInCode;

		CodeTableEntry* codeTableEntry = codeTable + i;
		hashCodeTable->insertItem(&(initialVals[i]), 1, codeTableEntry);
	}

	codeTable[i].string = NULL;
	codeTable[i].stringLen = 0;
	codeTable[i].code = clearCode;
	codeTable[i].bitsInCode = *bitsInCode;

	codeTable[i+1].string = NULL;
	codeTable[i+1].stringLen = 0;
	codeTable[i+1].code = clearCode + 1;
	codeTable[i+1].bitsInCode = *bitsInCode;
}






//
// find the number of significant bits in a word.  This isn't the most
// efficient way of doing things, but it will work for now.
//
static int numSignificantBits(unsigned int word) {
	int lastBit=0;

	for (int i=0; i < sizeof(unsigned int) * 8; i++) {
		if (word >> i) {
			lastBit = i+1;
		}
	}

	return lastBit;
}





//
// return a mask that will be all 1's for the number of bits specified.
// The mask will be shifted over by the specified number of bits
//
static unsigned char createMask(int numBits, int shift);



//
// this function will add the given code to the data bytes.
// It has a side effect of updating the currentDataBit variable.
//
// it will return the total number of valid bytes in this block that
// have been written.
//
// This function will also write the dataBytes out to disk if the number
// of data bytes goes over the maximum allowable block size (255)
//

static int setDataBytes(unsigned int code, int numSigBits,
				        unsigned char *dataBytes, int *currentDataBit,
						FILE* imageFile) {
	int currentByte = *currentDataBit / 8;
	int currentBitInByte = *currentDataBit % 8;
	int numBitsWritten = 0;

	int oldCurrentByte, oldCurrentBitInByte, oldCurrentDataBit;


	//
	// for gif images, we are guaranteed that the numSigBits should never
	// be above 12, So, a code should never span more than 3 bytes.
	//
	while (numBitsWritten < numSigBits) {
		int numBitsToGoInByte = 8 - currentBitInByte;
		int numBitsToWriteInByte = MIN(numSigBits-numBitsWritten, 
		                               numBitsToGoInByte);

		if (numBitsToGoInByte == 8) {
			dataBytes[currentByte] = 0;
		}

		unsigned char mask = createMask(numBitsToWriteInByte,
		                                currentBitInByte);

		dataBytes[currentByte] |= (unsigned char)(((code >> numBitsWritten) << currentBitInByte) & mask);

		
		oldCurrentDataBit = *currentDataBit;
		oldCurrentByte = currentByte;
		oldCurrentBitInByte = currentBitInByte;

		numBitsWritten += numBitsToGoInByte;
		*currentDataBit += numBitsToWriteInByte;
		currentByte = *currentDataBit / 8;
		currentBitInByte = *currentDataBit % 8;


		//
		// we need to check to see if we need to write out this block out
		// do disk.  If so, write it out.
		//
		if (currentBitInByte == 0 && 
			*currentDataBit + numSigBits-numBitsWritten + numSigBits+1 > 255*8) {

			writeBlockToFile(dataBytes, currentByte, imageFile);

			currentByte = 0;
			*currentDataBit = 0;
		}
	}


	if (currentBitInByte) {
		currentByte += 1;
	}

	return currentByte;
}






//
// return a mask that will be all 1's for the number of bits specified.
// The mask will be shifted over by the specified number of bits
//
static unsigned char createMask(int numBits, int shift) {
	static unsigned char table[] = { 0x01, 0x03, 0x07, 0x0f,
	                                 0x1f, 0x3f, 0x7f, 0xff };

	unsigned char mask=0;


	if (numBits > 0) {
		if (numBits <= 8) {
			mask = table[numBits-1];
		}

		mask = mask << shift;
	}


	return mask;
}







//
// see if the current prefix + the current pixel is in the table.
//
static int isInTable(unsigned char* prefix, int prefixLen, 
					 unsigned char currentPixel, 
					 OpenHashTable<CodeTableEntry*>* hashCodeTable,
					 int lastCodeTableEntry) {

	CodeTableEntry* codeTableEntry;

	unsigned char* item = new unsigned char[prefixLen+1];
	memcpy(item, prefix, prefixLen);
	item[prefixLen] = currentPixel;

	codeTableEntry = hashCodeTable->getItem(item, prefixLen+1);

	delete [] item;


	if (codeTableEntry) {
		return 1;
	}
	else {
		return 0;
	}
}





//
// return the code that is contained in the table.  The bitsInCode will
// get set to the number of bits in the code that are significant.
//
// if you use this function, the prefix must be in the table or else you 
// will most likely core or go into an infinite loop.  Use isInTable to
// see if the string is in the table.
//
static unsigned int lookUpCodeInTable(unsigned char* prefix, int prefixLen, 
									  OpenHashTable<CodeTableEntry*>* hashCodeTable,
									  int* bitsInCode) {
	CodeTableEntry* codeTableEntry;
	codeTableEntry = hashCodeTable->getItem(prefix, prefixLen);

	if (codeTableEntry) {
		*bitsInCode = codeTableEntry->bitsInCode;
		return codeTableEntry->code;
	}
	else {
		*bitsInCode = 0;
		return 0;
	}
}







//
// I don't need this function any more...
//
static unsigned int intCreateMask(int numBits) {
	static unsigned int table[] = { 
		0x00000001, 0x00000003, 0x00000007, 0x0000000f,
	    0x0000001f, 0x0000003f, 0x0000007f, 0x000000ff,
		0x000001ff, 0x000003ff, 0x000007ff, 0x00000fff,
		0x00001fff, 0x00003fff, 0x00007fff, 0x0000ffff,
		0x0001ffff, 0x0003ffff, 0x0007ffff, 0x000fffff,
		0x001fffff, 0x003fffff, 0x007fffff, 0x00ffffff,
		0x01ffffff, 0x03ffffff, 0x07ffffff, 0x0fffffff,
		0x1fffffff, 0x3fffffff, 0x7fffffff, 0xffffffff
	};

	unsigned int mask=0;


	if (numBits > 0) {
		if (numBits <= sizeof(unsigned int)*8) {
			mask = table[numBits-1];
		}
	}


	return mask;
}





//
// add the specified prefix to the table.  A new code will automatically
// be assigned.  lastCodeTableEntry will be incremented if necessary.
// If the code was successfully added, this function will return 1.  If the
// table is full, it will return 0.
//
static int addPrefixToCodeTable(unsigned char* prefix, int prefixLen, 
						        OpenHashTable<CodeTableEntry*>* hashCodeTable, 
								CodeTableEntry* codeTable,
						        int* lastCodeTableEntry) {
	if (*lastCodeTableEntry >= 0xfff) {
		return 0;
	}

	codeTable[*lastCodeTableEntry].string = prefix;
	codeTable[*lastCodeTableEntry].stringLen = prefixLen;


	unsigned int prevCode = codeTable[*lastCodeTableEntry-1].code;
	int prevBitsInCode = codeTable[*lastCodeTableEntry-1].bitsInCode;

	codeTable[*lastCodeTableEntry].code = prevCode + 1;

	if (codeTable[*lastCodeTableEntry].code >> prevBitsInCode) {
		//codeTable[*lastCodeTableEntry].bitsInCode = prevBitsInCode+1;
		//
		// we need more space, so set each entry in the table to contain
		// one more significant bit
		//
		for (int i=0; i <= *lastCodeTableEntry; i++) {
			codeTable[i].bitsInCode = prevBitsInCode+1;
		}
	}
	else {
		codeTable[*lastCodeTableEntry].bitsInCode = prevBitsInCode;
	}


	if (codeTable[*lastCodeTableEntry].bitsInCode > 12) {
		return 0;
	}


	//
	// besides entering the new code into the orderded code table, also
	// enter it into the hash table for easy searching.
	//
	CodeTableEntry* codeTableEntry = &(codeTable[*lastCodeTableEntry]);
	hashCodeTable->insertItem(prefix, prefixLen, codeTableEntry);


	*lastCodeTableEntry += 1;


	return 1;
}







//
// write the specified block out to the file as specified by the gif format
// 
static int writeBlockToFile(unsigned char* dataBytes, 
							unsigned char blockByteCount,
							FILE* imageFile) {

	int status = fwrite(&blockByteCount, 1, 1, imageFile);

	if (status != 1) {
		return 0;
	}


	if (blockByteCount && !dataBytes) {
		return 0;
	}
	else if (blockByteCount) {
		status = fwrite(dataBytes, 1, blockByteCount, imageFile);

		if (status != blockByteCount) {
			return 0;
		}
	}

	return 1;
}


