// This is ImageFileRGBLoad.C
//
// Copyright 1997, 1998     Mark Manca
//

// This file contains the code to load an rgb image

#ifdef WIN32
#include <windows.h>
#endif

#include <GL/gl.h>
#include <string.h>
#include <stdio.h>

#include "ImageFile.h"
#include "rgbFileHeader.h"

#ifdef LITTLE_ENDIAN
#include "byteSwap.h"
#endif


//
// These _readRGBHeader functions assumes that the file to be 
// read is already open.
//
int ImageFile::_readRGBHeader() {
	rgbFileHeader fileInfo;

	return this->_readRGBHeader(&fileInfo);
}


int ImageFile::_readRGBHeader(rgbFileHeader *fileInfo) {

	int status;

	if (!__imageFile || !__isOpen) {
		return 0;
	}


	//
	// get to the beginning of the file so we can read the 
	// header into the struct
	//
	status = fseek(__imageFile, 0, SEEK_SET);
	if (status) {
		return 0;			// an error occured
	}


	//
	// read in the header into the data structure
	//
	status = fread((char *)fileInfo, 1, sizeof(rgbFileHeader), __imageFile);

	// should we use ferror here??
	if (status != sizeof(rgbFileHeader)) {
		return 0;
	}


	#ifdef LITTLE_ENDIAN	// need to swap bytes
	fileInfo->magicNumber   = byteSwap(fileInfo->magicNumber);
	fileInfo->numDimensions = byteSwap(fileInfo->numDimensions);
	fileInfo->xSize         = byteSwap(fileInfo->xSize);
	fileInfo->ySize         = byteSwap(fileInfo->ySize);
	fileInfo->zSize         = byteSwap(fileInfo->zSize);
	fileInfo->minPixelValue = halfwordAndByteSwap(fileInfo->minPixelValue);
	fileInfo->maxPixelValue = halfwordAndByteSwap(fileInfo->maxPixelValue);
	fileInfo->colormapID    = halfwordAndByteSwap(fileInfo->colormapID);
	#endif // WIN32


	#ifdef _DEBUG
	printf("storageFormat: %d\n", (int)fileInfo->storageFormat);
	printf("bytesPerChannel: %d\n", (int)fileInfo->bytesPerChannel);
	printf("numDimensions: %d\n", fileInfo->numDimensions);
	printf("xSize: %d\nySize: %d\nzSize: %d\n", 
	       fileInfo->xSize, fileInfo->ySize, fileInfo->zSize);
	printf("minPixelValue: %d\n", fileInfo->minPixelValue);
	printf("maxPixelValue: %d\n", fileInfo->maxPixelValue);
	fflush(stdout);
	#endif // _DEBUG


	//
	// rgb files provide the user a way to name the picture inside of
	// the file itself. So, set the image name to the name that was in 
	// the file.
	//
	this->setName(fileInfo->imageName);


	//
	// rgb files support both 8 bits per pixel per color or 16 bits
	// per pixel per color. So, find the correct format that the 
	// data is in and set up the class accordingly.
	//
	switch (fileInfo->bytesPerChannel) {
		case 1:
			_imageDataType = GL_BYTE;
			break;
		case 2:
			_imageDataType = GL_SHORT;
			break;
		default:
			fprintf(stderr, "%s: bad value in bytes per channel field\n", __filename);
	}


	//
	// change some data types around for efficiency if possible
	//
	if (fileInfo->minPixelValue >= 0  && fileInfo->maxPixelValue >= 0) {
		if (_imageDataType == GL_BYTE) {
			_imageDataType = GL_UNSIGNED_BYTE;
		}
		else {
			_imageDataType = GL_UNSIGNED_SHORT;
		}
	}


	//
	// a file can be grayscale, 3 channel color, or 4 channel color
	// (3 color channels + alpha). So, figure out how many channels 
	// are stored in this image.
	//
	if (_colormapData) {
		delete [] _colormapData;
		_colormapData = NULL;
	}
	if (_imageData) {
		delete [] _imageData;
		_imageData = NULL;
	}


	switch (fileInfo->numDimensions) {
		//
		// first two cases are wierd instances of black and white images
		//
		case 1:
			_width = fileInfo->xSize;
			_height = 1;
			_numChannels = 1;
			_imageType = GL_COLOR_INDEX;		// this is really a bw image
			_colormapSize = 256;
			_colormapDataType = GL_UNSIGNED_BYTE;
			break;

		case 2:
			_width = fileInfo->xSize;
			_height = fileInfo->ySize;
			_imageType = GL_COLOR_INDEX;		// this is really a bw image
			_numChannels = 1;
			_colormapSize = 256;
			_colormapDataType = GL_UNSIGNED_BYTE;
			break;

		//
		// This is the most common case. Even if we get a black and
		// white image, expand the image out to full color in memory
		// to make things simpler later on down the line...
		//
		case 3:
			_width = fileInfo->xSize;
			_height = fileInfo->ySize;
			_numChannels = fileInfo->zSize;
			_colormapSize = 0;
			_colormapDataType = GL_UNSIGNED_BYTE;

			switch (_numChannels) {
				case 1:
					_imageType = GL_COLOR_INDEX;	// this is really a bw image
					break;
				case 3:
					_imageType = GL_RGB;
					break;
				case 4:
					_imageType = GL_RGBA;
					break;
				default:
					fprintf(stderr, "%s: bad value in zSize field\n", __filename);
			}
	
			//
			// if the colormapID is set we have a colormapped image.
			// As indicated in the file format spec, this format
			// is outdated. Include it for backwards compatibility.
			//
			if (fileInfo->colormapID != 0) {
				_imageType = GL_COLOR_INDEX;
			}
			
			break;

		default:
			fprintf(stderr, "%s: bad value in dimensions field\n", __filename);
	}



	return 1;
}




//
// This function will read in the data from an rgb file into the image class
//
int ImageFile::_readRGBFile() {
	rgbFileHeader fileInfo;
	int status;


	//
	// read in the data header. In some cases we might already have
	// done this. If that is the case, oh well, we are not wasting too
	// much time here by reading it in again.
	//
	if (!this->_readRGBHeader(&fileInfo)) {
		return 0;
	}


	int rowDataSize;
	void *rowData;

	//
	// allocate the memory that will be used to store the image
	//
	_allocateImageData();


	//
	// although this call gets called, it won't really do anything unless
	// we are loading in a black and white image
	//
	_allocateColormapData();


	//
	// allocate memory that will be used to read in the image
	//
	// mutiply width * 2 everywhere to make sure that we have enough
	// room when we are reading the the run length encoded data. width *2
	// should be the maximum amount that an rle encoded line should be.
	//
	if (_imageDataType == GL_BYTE) {
		//
		// These values are used as we read in the image a row at a time
		//
		rowDataSize = _width;
		rowData = new GLbyte[_width * 2]; 
	}
	else if (_imageDataType == GL_UNSIGNED_BYTE) {
		rowDataSize = _width;
		rowData = new GLubyte[_width * 2];
	}	
	else if (_imageDataType == GL_SHORT) {
		rowDataSize = 2 * _width;
		rowData = new GLshort[_width * 2];
	}
	else {		// we've got unsigned shorts...
		rowDataSize = 2 * _width;
		rowData = new GLushort[_width * 2];
	}



	int i, j, k, l;
	int count;
	int heightOffset, widthOffset;
	int imageRowDataSize = _width*_numChannels + _extraRowWidth;

	//
	// now that prep work is done, begin the actual loading of the file
	//

	//
	// The rgb file format supports two encoding schemes: plain and
	// Run Length Encoding (rle). Run length encoding reduces the filesize
	// by keeping track of a count of how many adjacent pixels have the
	// same color. It only compresses a row at a time. For more info,
	// see the spec.
	//
	// If no rle, reading the file is reletavely easy.
	//
	if (fileInfo.storageFormat == 0) {

		//
		// an rgb file stores each channel separately. For a full
		// color image, first it contains the red pixels, then the
		// green, and then the blue. If alpha is present, those values
		// will follow the blue.
		//
		for (i=0; i < _numChannels; i++) {
			heightOffset = 0;
			for (j=0; j < _height; j++) {

				//
				// read in a row from the file
				//
				status = fread(rowData, 1, rowDataSize, __imageFile);
				if (status != rowDataSize) {
					if (feof(__imageFile)) {
						fprintf(stderr, "%s: premature eof\n", __filename);
					}
					if (ferror(__imageFile)) {
						fprintf(stderr, "%s: error on read\n", __filename);
					}
					return 0;
				}


				widthOffset = _width * _numChannels;

				//
				// now that we are in the most inner most loop,
				// do some convoluted things so we can be as 
				// efficient as possible.
				//
				// first of all, duplicate some code by putting
				// the if statements outside of the loop.
				//

				if (_imageDataType == GL_BYTE) {
					//
					// for efficency, go backwards and 
					// compare with 0.
					//
					for (k=_width-1; k >= 0; k--) {
						widthOffset -= _numChannels;

						((GLbyte*)_imageData)[heightOffset+widthOffset+i] = 
							((GLbyte *)rowData)[k];
					}
				}
				else if (_imageDataType == GL_UNSIGNED_BYTE) {
					for (k=_width-1; k >= 0; k--) {
						widthOffset -= _numChannels;

						((GLubyte*)_imageData)[heightOffset+widthOffset+i] = 
							((GLubyte *)rowData)[k];
					}
				}
				else if (_imageDataType == GL_SHORT) {
					for (k=_width-1; k >= 0; k--) {
						widthOffset -= _numChannels;

						((GLshort *)_imageData)[heightOffset+widthOffset+i] = 
							((GLshort *)rowData)[k];

						#ifdef LITTLE_ENDIAN // need to byte swap
							((GLshort *)_imageData)[heightOffset+widthOffset+i] = 
								byteSwap(((GLshort *)rowData)[k]);
						#endif
					}
				}
				else {
					for (k=_width-1; k >= 0; k--) {
						widthOffset -= _numChannels;

						((GLushort *)_imageData)[heightOffset+widthOffset+i] = 
							((GLushort *)rowData)[k];

						#ifdef LITTLE_ENDIAN // need to byte swap
							((GLushort *)_imageData)[heightOffset+widthOffset+i] = 
								byteSwap(((GLushort *)rowData)[k]);
						#endif
					}
				}

				//
				// calculate the byte offset into memory. The
				// lower left hand corner of the image is stored
				// in the first byte (ala OpenGL)
				//

				heightOffset += imageRowDataSize;
			}
		}
	}
	else if (fileInfo.storageFormat == 1) {    // file is rle encoded
		unsigned int *startTable;
		unsigned int *lengthTable;
		int tablen = fileInfo.ySize * fileInfo.zSize;
		int rleOffset;
		int rleLength;
		int longSize = sizeof(long);
			
		startTable = new unsigned int[tablen];
		lengthTable = new unsigned int[tablen];

		// read in the necessary data from the rle tables
		status = fread((char*)startTable, 1, longSize * tablen, __imageFile);
		if (status != longSize * tablen) {
			return 0;
		}
		status = fread((char*)lengthTable, 1, longSize * tablen, __imageFile);
		if (status != longSize * tablen) {
			return 0;
		}

		#ifdef LITTLE_ENDIAN	// need to byteswap tables
			for (i=tablen-1; i >= 0; i--) {
				startTable[i] = halfwordAndByteSwap(startTable[i]);
				lengthTable[i] = halfwordAndByteSwap(lengthTable[i]);
			}
		#endif


		// begin the process of actually reading in the file
		for (i=0; i < _numChannels; i++) {
			heightOffset = 0;

			for (j=0; j < _height; j++) {

				rleOffset = startTable[i*_height + j];
				rleLength = lengthTable[i*_height + j];

				// get to proper place in file
				status = fseek(__imageFile, rleOffset, SEEK_SET);
				if (status) {
					return 0;
				}


				// will rowData be big enough???
				status = fread(rowData, 1, rleLength, __imageFile);
				if (status != rleLength) {
					return 0;
				}
		
				#ifdef LITTLE_ENDIAN // need to swap bytes we just read
				// swapping still needs to be implemented for rle data
				// have to do this only if we are dealing w/shorts
				#endif

				k=0;
				l=0;
				// while count != 0
				// we only handle byte and unsigned bytes right now...
				while((((_imageDataType == GL_BYTE || _imageDataType == GL_UNSIGNED_BYTE) ?
					(((unsigned char *)rowData)[k] & (unsigned char)0x7f) != 0 : (((unsigned short *)rowData)[k] & 0x007f) != 0))
					&& k < rleLength) {
				
					if (_imageDataType == GL_BYTE || _imageDataType == GL_UNSIGNED_BYTE) {
						count = ((unsigned char *)rowData)[k] & (unsigned char)0x7f;

						// copy count bytes to the image buffer
						if (((unsigned char *)rowData)[k] & (unsigned char)0x80) {
							k++;

							if (_imageDataType == GL_BYTE) {
								for (; count > 0; count--, l++, k++) {
									widthOffset = l * _numChannels;
									((GLbyte*)_imageData)[heightOffset+widthOffset+i] = ((GLbyte *)rowData)[k];
								}
							}
							else {      // _imageDataType == GL_UNSIGNED_BYTE
								for (; count > 0; count--, l++, k++) {
									widthOffset = l * _numChannels;
									((GLubyte*)_imageData)[heightOffset+widthOffset+i] = ((GLubyte *)rowData)[k];
								}
							}

						}
						else {	// copy the next byte count times
							k++;

							if (_imageDataType == GL_BYTE) {
								for (; count > 0; count--, l++)	{
									widthOffset = l * _numChannels;
									((GLbyte*)_imageData)[heightOffset+widthOffset+i] = ((GLbyte *)rowData)[k];
								}
							}
							else {       // _imageDataType == GL_UNSIGNED_BYTE
								for (; count > 0; count--, l++)	{
									widthOffset = l * _numChannels;
									((GLubyte*)_imageData)[heightOffset+widthOffset+i] = ((GLubyte *)rowData)[k];
								}
							}

							k++;	// advance to next count var
						}
					}
#if 0
					// if data is stored in shorts
					else {
						count = ((short *)rowData)[k] & 0x7f;

						// copy count bytes to the image buffer
						if (((short *)rowData)[k] & 0x80) {
							k++;

							if (_imageDataType == GL_SHORT) {
								for (; count > 0; count--, l++, k++) {
									widthOffset = l * _numChannels;
									((GLshort*)_imageData)[heightOffset+widthOffset+i] = ((GLshort *)rowData)[k];
								}
							}
							else {      // _imageDataType == GL_UNSIGNED_SHORT
								for (; count > 0; count--, l++, k++) {
									widthOffset = l * _numChannels;
									((GLushort*)_imageData)[heightOffset+widthOffset+i] = ((GLushort *)rowData)[k];
								}
							}
						}
						else {    // copy the next byte count times
							k++;

							if (_imageDataType == GL_SHORT) {
								for (; count > 0; count--, l++) {
									widthOffset = l * _numChannels;
									((GLshort*)_imageData)[heightOffset+widthOffset+i] = ((GLshort *)rowData)[k];
								}
							}
							else {       // _imageDataType == GL_UNSIGNED_SHORT
								for (; count > 0; count--, l++) {
									widthOffset = l * _numChannels;
									((GLushort*)_imageData)[heightOffset+widthOffset+i] = ((GLushort *)rowData)[k];
								}
							}
							k++;    // advance to next count var
						}
					}
#endif
				}

				#if 0
				// If display while read is set, display a row
				if (__drawWhileRead) {
					this->drawRow(j);
				}
				#endif

				heightOffset += imageRowDataSize;
			}
		}

		delete [] startTable;
		delete [] lengthTable;
	}
	else {
		fprintf(stderr, "%s: invalid encoding scheme\n", __filename);
	}



#if 0
	//
	// if we've got black and white data, assign the correct color pallete
	// 

	if (_numChannels == 1 && _imageDataType == GL_COLOR_INDEX) {
		//if (
		//for (i=_colormapData
		for (i=0; i < _height; i++) {
			heightOffset = i * _width * _numChannels 
			               + j*_extraRowWidth;

			for (j=_width-1; j >= 0; j--) {
				widthOffset = j * _numChannels;

				if (_imageDataType == GL_BYTE 
					|| _imageDataType == GL_UNSIGNED_BYTE) {
					
					((char*)_imageData)[heightOffset+widthOffset+1] = 
							((char*)_imageData)[heightOffset+widthOffset];
					((char*)_imageData)[heightOffset+widthOffset+2] = 
							((char*)_imageData)[heightOffset+widthOffset];
				}
				else {
					((short*)_imageData)[heightOffset+widthOffset+1] = 
							((short*)_imageData)[heightOffset+widthOffset];
					((short*)_imageData)[heightOffset+widthOffset+2] = 
							((short*)_imageData)[heightOffset+widthOffset];
				}
			}
		}
	}
#endif


	delete [] rowData;

	
	return 1;
}


